/* All code copyright Paul Callahan 1996  (callahan@inf.ethz.ch) */

#include <stdio.h>
#include <limits.h>
#include "gen.h"
#include <malloc.h>
#include <assert.h>

static int rules[]={
                   0, 0,
                   0, 0,
                   0, 1,
                   1, 1,
                   0, 0,
                   0, 0, 
                   0, 0, 
                   0, 0, 
                   0, 0 };

int spreadColor[5]={ 0x0000, 0x0001, 0x0010, 0x0100, 0x1000 };

static int neighborhoodTable[1<<9];
static int initializedTable= 0;


/* maintain history of cell populations for filtering low-period oscillations */
static int pop[POPHISTORYSIZE];
static int repetitions[POPHISTORYSIZE];
int iPop;

void resizeIfNeeded(LifeList *cells, int n) {

  if (n > cells->maxsize) {
    Cell *tmp;     
    int newsize=2*n;

    if (cells->maxsize==0) {
      cells->cellList=(Cell *)calloc(sizeof(Cell), newsize);
      cells->scratch1=(Cell *)calloc(sizeof(Cell), newsize);
      cells->scratch2=(Cell *)calloc(sizeof(Cell), newsize);
      cells->neighborhoods=(Cell *)calloc(sizeof(Cell), newsize);
    }
    else {
      int i; 

      tmp=(Cell *)calloc(sizeof(Cell), newsize);
      for (i=0; i<cells->maxsize; i++) tmp[i]=cells->cellList[i];
      free(cells->cellList);
      cells->cellList=tmp;

      tmp=(Cell *)calloc(sizeof(Cell), newsize);
      for (i=0; i<cells->maxsize; i++) tmp[i]=cells->scratch1[i];
      free(cells->scratch1);
      cells->scratch1=tmp;

      tmp=(Cell *)calloc(sizeof(Cell), newsize);
      for (i=0; i<cells->maxsize; i++) tmp[i]=cells->scratch2[i];
      free(cells->scratch2);
      cells->scratch2=tmp;

      tmp=(Cell *)calloc(sizeof(Cell), newsize);
      for (i=0; i<cells->maxsize; i++) tmp[i]=cells->neighborhoods[i];
      free(cells->neighborhoods);
      cells->neighborhoods=tmp;

    }
    cells->maxsize=newsize;
/*
    printf("newsize: %d\n", cells->maxsize);
*/
  }
}

int nextGen(Cell *counts, int ncounts, Cell *cellList) {

   int i, ncells;

   ncells=0;

   for (i=0; i<ncounts; i++) {
     if (rules[counts[i].value]) {
       cellList[ncells].position= counts[i].position;
       cellList[ncells++].value=2;
     }
   }

   return ncells;
 }

int nextGenFromTable(Cell *neighborhoods, int nneighborhoods, Cell *cellList) {

   int i, ncells;

   ncells=0;

   for (i=0; i<nneighborhoods; i++) {
     if (neighborhoodTable[neighborhoods[i].value]) {
       cellList[ncells].position= neighborhoods[i].position;
       cellList[ncells++].value=1;
     }
   }

   return ncells;
 }

int sumNeighbors(LifeList *cells) {

  int n1, n2, n3;

  resizeIfNeeded(cells, 9*cells->ncells);
  n3=sumAllNeighbors(cells->cellList, cells->ncells, cells->neighborhoods);

  return n3;
}


void makeRowMajor(LifeList *cells) {

   sortRowMajor(cells->cellList, cells->ncells);
   cells->ncells= uniqCells(cells->cellList, cells->ncells);

}

void oscCheck(int n) {

   int i;

  /* low period oscillation check */
  for (i=1; i<=POPHISTORYSIZE; i++) {
     if (pop[(iPop+POPHISTORYSIZE-i)%POPHISTORYSIZE]==n) {
        repetitions[i-1]++;
     } else {
        repetitions[i-1]=0;
     } 
  }
  pop[iPop]= n;
  iPop=(iPop+1)%POPHISTORYSIZE;


}

void initNeighborhoodTable() {

  int i, b0, b1, b2, b3, b4, b5, b6, b7, b8;

  for (i=0; i<(1<<9); i++) {
     b0=    i  & 1;
     b1=(i>>1) & 1; 
     b2=(i>>2) & 1; 
     b3=(i>>3) & 1; 
     b4=(i>>4) & 1; 
     b5=(i>>5) & 1; 
     b6=(i>>6) & 1; 
     b7=(i>>7) & 1; 
     b8=(i>>8) & 1; 

     neighborhoodTable[i]=rules[b4 +2*(b0+b1+b2+b3+b5+b6+b7+b8)];
  } 

}


void generate(LifeList *cells) {

  int i,n;

  if (!initializedTable) {
     initNeighborhoodTable();
     initializedTable= 1;
  }

  resizeIfNeeded(cells, 9*cells->ncells);

  n=findNeighborhoods(cells->cellList, cells->ncells, cells->neighborhoods);

  cells->ncells= nextGenFromTable(cells->neighborhoods, n, cells->cellList);

  oscCheck(cells->ncells);
}


History makeHistory(LifeList *cells, int ngenerations) {

    int i;
    History hist;

    hist.cellList= (Cell **)calloc(sizeof(Cell *), ngenerations);
    hist.ncells= (int *)calloc(sizeof(int), ngenerations);
    hist.ngenerations= ngenerations;

    setValues(cells->cellList, cells->ncells, 1);

    for (i=0; i<ngenerations; i++) {
       hist.cellList[i]=(Cell *)calloc(sizeof(Cell), cells->ncells);

       hist.ncells[i]=
          copyList(cells->cellList, cells->ncells, hist.cellList[i], 0);

       generate(cells);
       
    }

    return hist;
}

void freeHistory(History hist) {
    int i;

    for (i=0; i<hist.ngenerations; i++) {
        free(hist.cellList[i]);
    }
    free(hist.cellList);
    free(hist.ncells);
}

void dumpHistory(History hist) {

    int i;

    for (i=0; i<hist.ngenerations; i++) {
       printf("generation: %3d cellcount: %3d\n", i, hist.ncells[i]);
       dumpCells(hist.cellList[i], hist.ncells[i]);
    }
}

void getGeneration(LifeList *cells, History hist, int gen) {

    assert(gen < hist.ngenerations);

    resizeIfNeeded(cells, hist.ncells[gen]);

    cells->ncells=
        copyList(hist.cellList[gen], hist.ncells[gen], cells->cellList, 0);

}

void initPopHistory() {

  int i;

  for (i=0; i<POPHISTORYSIZE; i++) { 
     pop[i]= -1;
     repetitions[i]= 0;     
  }

  iPop= 0;
  
}

void longestOscillation(int *period, int *repetition) {
int i;

   *repetition= -1; 

   for (i=0; i<POPHISTORYSIZE; i++) {   
       if (repetitions[i] > *repetition) {
          *period = i+1;
          *repetition= repetitions[i]; 
       }
   }

}
